@@ -19,14 +19,36 @@
 #include <unistd.h>
 #include <wchar.h>
 
-#define LEN(a)              sizeof((a))/sizeof((a)[0])
-#define PAD_TRUNCATE_SYMBOL "\xe2\x80\xa6" /* symbol: "ellipsis" */
-#define SCROLLBAR_SYMBOL    "\xe2\x94\x82" /* symbol: "light vertical" */
+/* Allow to lazyload items when a file is specified? This saves memory but
+   increases some latency when seeking items. It also causes issues if the
+   feed is changed while having the UI open (and offsets are changed). */
+/*#define LAZYLOAD 1*/
+
+#define LEN(a) sizeof((a))/sizeof((a)[0])
+
+#define PAD_TRUNCATE_SYMBOL    "\xe2\x80\xa6" /* symbol: "ellipsis" */
+#define SCROLLBAR_SYMBOL_BAR   "\xe2\x94\x82" /* symbol: "light vertical" */
+#define SCROLLBAR_SYMBOL_TICK  " "
+
+/* See the README for some color theme examples. */
+#define THEME_ITEM_NORMAL()           do {                            } while(0)
+#define THEME_ITEM_FOCUS()            do {                            } while(0)
+#define THEME_ITEM_BOLD()             do { attrmode(ATTR_BOLD_ON);    } while(0)
+#define THEME_ITEM_SELECTED()         do { attrmode(ATTR_REVERSE_ON); } while(0)
+#define THEME_SCROLLBAR_FOCUS()       do {                            } while(0)
+#define THEME_SCROLLBAR_NORMAL()      do { attrmode(ATTR_FAINT_ON);   } while(0)
+#define THEME_SCROLLBAR_TICK_FOCUS()  do { attrmode(ATTR_REVERSE_ON); } while(0)
+#define THEME_SCROLLBAR_TICK_NORMAL() do { attrmode(ATTR_REVERSE_ON); } while(0)
+#define THEME_STATUSBAR()             do { attrmode(ATTR_REVERSE_ON); } while(0)
+#define THEME_INPUT_LABEL()           do { attrmode(ATTR_REVERSE_ON); } while(0)
+#define THEME_INPUT_NORMAL()          do {                            } while(0)
+
+static char *plumber = "xdg-open"; /* environment variable: $SFEED_PLUMBER */
+static char *piper = "sfeed_content"; /* environment variable: $SFEED_PIPER */
+static char *yanker = "xclip -r"; /* environment variable: $SFEED_YANKER */
 
 enum {
-	ATTR_RESET = 0,
-	ATTR_BOLD_ON = 1, ATTR_FAINT_ON = 2,
-	ATTR_REVERSE_ON = 7, ATTR_REVERSE_OFF = 27,
+	ATTR_RESET = 0,	ATTR_BOLD_ON = 1, ATTR_FAINT_ON = 2, ATTR_REVERSE_ON = 7
 };
 
 enum Pane { PaneFeeds, PaneItems, PaneLast };
@@ -37,29 +59,29 @@ enum {
 };
 
 struct win {
-	int width;
-	int height;
-	int dirty;
+	int width; /* absolute width of the window */
+	int height; /* absolute height of the window */
+	int dirty; /* needs draw update: clears screen */
 };
 
 struct row {
-	char *text;
+	char *text; /* text string, optional if using row_format() callback */
 	int bold;
-	void *data;
+	void *data; /* data binding */
 };
 
 struct pane {
 	int x; /* absolute x position on the screen */
 	int y; /* absolute y position on the screen */
-	int width; /* absolute width of the window */
-	int height; /* absolute height of the window */
+	int width; /* absolute width of the pane */
+	int height; /* absolute height of the pane */
 	off_t pos; /* focused row position */
 	struct row *rows;
 	size_t nrows; /* total amount of rows */
 	int focused; /* has focus or not */
 	int hidden; /* is visible or not */
 	int dirty; /* needs draw update */
-
+	/* (optional) callback functions */
 	struct row *(*row_get)(struct pane *, off_t pos);
 	char *(*row_format)(struct pane *, struct row *);
 	int (*row_match)(struct pane *, struct row *, const char *);
@@ -87,7 +109,6 @@ struct statusbar {
 
 /* /UI */
 
-/* feed info */
 struct feed {
 	char         *name;     /* feed name */
 	char         *path;     /* path to feed or NULL for stdin */
@@ -142,15 +163,26 @@ static size_t nurls;
 
 volatile sig_atomic_t sigstate = 0;
 
-/* config */
+int
+ttywritef(const char *fmt, ...)
+{
+	va_list ap;
+	int n;
 
-/* Allow to lazyload items when a file is specified? This saves memory but
-   increases some latency when seeking items. It also causes issues if the
-   feed is changed while having the UI open (and offsets are changed). */
-/*#define LAZYLOAD 1 */
+	va_start(ap, fmt);
+	n = vdprintf(1, fmt, ap);
+	va_end(ap);
 
-static char *plumber = "xdg-open"; /* environment variable: $SFEED_PLUMBER */
-static char *piper = "less"; /* environment variable: $SFEED_PIPER */
+	return n;
+}
+
+int
+ttywrite(const char *s)
+{
+	if (!s)
+		return 0; /* for tparm() returning NULL */
+	return write(1, s, strlen(s));
+}
 
 /* like BSD err(), but calls cleanup() and _exit(). */
 void
@@ -163,12 +195,12 @@ err(int code, const char *fmt, ...)
 	cleanup();
 
 	va_start(ap, fmt);
-	vfprintf(stderr, fmt, ap);
+	vdprintf(2, fmt, ap);
 	va_end(ap);
 
-	fputs(": ", stderr);
-	errno = saved_errno;
-	perror(NULL);
+	if (saved_errno)
+		dprintf(2, ": %s", strerror(saved_errno));
+	write(2, "\n", 1);
 
 	_exit(code);
 }
@@ -282,7 +314,7 @@ colw(const char *s)
 	return col;
 }
 
-/* format `len' columns of characters. If string is shorter pad the rest
+/* Format `len' columns of characters. If string is shorter pad the rest
  * with characters `pad`. */
 int
 utf8pad(char *buf, size_t bufsiz, const char *s, size_t len, int pad)
@@ -332,14 +364,14 @@ void
 printpad(const char *s, int width)
 {
 	char buf[1024];
-	utf8pad(buf, sizeof(buf), s, width, ' ');
-	fputs(buf, stdout);
+	if (utf8pad(buf, sizeof(buf), s, width, ' ') != -1)
+		ttywrite(buf);
 }
 
 void
 resettitle(void)
 {
-	fputs("\x1b""c", stdout); /* rs1: reset title and state */
+	ttywrite("\x1b""c"); /* rs1: reset title and state */
 }
 
 void
@@ -352,48 +384,48 @@ updatetitle(void)
 		totalnew += feeds[i].totalnew;
 		total += feeds[i].total;
 	}
-	printf("\x1b]2;(%lu/%lu) - sfeed_curses\x1b\\", totalnew, total);
+	ttywritef("\x1b]2;(%lu/%lu) - sfeed_curses\x1b\\", totalnew, total);
 }
 
 void
 appmode(int on)
 {
-	/*fputs(on ? "\x1b[?1049h" : "\x1b[?1049l", stdout);*/ /* smcup, rmcup */
-	putp(tparm(on ? enter_ca_mode : exit_ca_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	/*ttywrite(on ? "\x1b[?1049h" : "\x1b[?1049l");*/ /* smcup, rmcup */
+	ttywrite(tparm(on ? enter_ca_mode : exit_ca_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
 mousemode(int on)
 {
-	fputs(on ? "\x1b[?1000h" : "\x1b[?1000l", stdout); /* xterm mouse mode */
+	ttywrite(on ? "\x1b[?1000h" : "\x1b[?1000l"); /* xterm mouse mode */
 }
 
 void
 cursormode(int on)
 {
-	/*fputs(on ? "\x1b[?25h" : "\x1b[?25l", stdout);*/ /* DECTCEM (in)Visible cursor */
-	putp(tparm(on ? cursor_visible : cursor_invisible, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	/*ttywrite(on ? "\x1b[?25h" : "\x1b[?25l");*/ /* DECTCEM (in)Visible cursor */
+	ttywrite(tparm(on ? cursor_normal : cursor_invisible, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
 cursorsave(void)
 {
-	/*fputs("\x1b""7", stdout);*/
-	putp(tparm(save_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	/*ttywrite("\x1b""7");*/
+	ttywrite(tparm(save_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
 cursorrestore(void)
 {
-	/*fputs("\x1b""8", stdout);*/
-	putp(tparm(restore_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	/*ttywrite("\x1b""8");*/
+	ttywrite(tparm(restore_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
 cursormove(int x, int y)
 {
-	/*printf("\x1b[%d;%dH", y + 1, x + 1);*/
-	putp(tparm(cursor_address, y, x, 0, 0, 0, 0, 0, 0, 0));
+	/*ttywritef("\x1b[%d;%dH", y + 1, x + 1);*/
+	ttywrite(tparm(cursor_address, y, x, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
@@ -401,30 +433,29 @@ attrmode(int mode)
 {
 	char *p;
 
-	/*printf("\x1b[%dm", mode);*/
+	/*ttywritef("\x1b[%dm", mode);*/
 	switch (mode) {
 	case ATTR_RESET: p = exit_attribute_mode; break;
 	case ATTR_BOLD_ON: p = enter_bold_mode; break;
 	case ATTR_FAINT_ON: p = enter_dim_mode; break;
 	case ATTR_REVERSE_ON: p = enter_standout_mode; break;
-	case ATTR_REVERSE_OFF: p = exit_standout_mode; break;
 	default: return;
 	}
-	putp(tparm(p, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	ttywrite(tparm(p, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
 cleareol(void)
 {
-	/*fputs("\x1b[K", stdout);*/
-	putp(tparm(clr_eol, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	/*ttywrite("\x1b[K");*/
+	ttywrite(tparm(clr_eol, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
 clearscreen(void)
 {
-	/*fputs("\x1b[H\x1b[2J", stdout);*/
-	putp(tparm(clear_screen, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	/*ttywrite("\x1b[H\x1b[2J");*/
+	ttywrite(tparm(clear_screen, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
@@ -447,7 +478,6 @@ cleanup(void)
 		mousemode(0);
 
 	resettitle();
-	fflush(stdout);
 
 	sigemptyset(&sa.sa_mask);
 	sa.sa_flags = SA_RESTART; /* require BSD signal semantics */
@@ -501,7 +531,6 @@ init(void)
 		mousemode(usemouse);
 
 	updategeom();
-	fflush(stdout);
 
 	sigemptyset(&sa.sa_mask);
 	sa.sa_flags = SA_RESTART; /* require BSD signal semantics */
@@ -514,11 +543,12 @@ init(void)
 	needcleanup = 1;
 }
 
-/* pipe item to a program, if `wantoutput` is set then cleanup and restore the
-   terminal attribute settings, if not set then don't do that and also ignore
-   stdout and stderr. */
+/* Pipe item line or item field to a program.
+   If `field` is -1 then pipe the TSV line, else a specified field.
+   if `wantoutput` is 1 then cleanup and restore the tty,
+   if 0 then don't do that and also write stdout and stderr to /dev/null. */
 void
-pipeitem(const char *cmd, struct item *item, int wantoutput)
+pipeitem(const char *cmd, struct item *item, int field, int wantoutput)
 {
 	FILE *fp;
 	int i, pid, wpid, status;
@@ -538,10 +568,14 @@ pipeitem(const char *cmd, struct item *item, int wantoutput)
 		errno = 0;
 		if (!(fp = popen(cmd, "w")))
 			err(1, "popen");
-		for (i = 0; i < FieldLast; i++) {
-			if (i)
-				fputc('\t', fp);
-			fputs(item->fields[i], fp);
+		if (field == -1) {
+			for (i = 0; i < FieldLast; i++) {
+				if (i)
+					fputc('\t', fp);
+				fputs(item->fields[i], fp);
+			}
+		} else {
+			fputs(item->fields[field], fp);
 		}
 		fputc('\n', fp);
 		status = pclose(fp);
@@ -560,7 +594,7 @@ pipeitem(const char *cmd, struct item *item, int wantoutput)
 }
 
 void
-plumb(const char *cmd, char *url)
+forkexec(char *argv[])
 {
 	switch (fork()) {
 	case -1:
@@ -568,7 +602,7 @@ plumb(const char *cmd, char *url)
 	case 0:
 		dup2(devnullfd, 1);
 		dup2(devnullfd, 2);
-		if (execlp(cmd, cmd, url, NULL) < 0)
+		if (execvp(argv[0], argv) == -1)
 			_exit(1);
 	}
 }
@@ -604,48 +638,32 @@ pane_row_match(struct pane *p, struct row *row, const char *s)
 }
 
 void
-pane_row_draw(struct pane *p, off_t pos)
+pane_row_draw(struct pane *p, off_t pos, int selected)
 {
 	struct row *row;
-	int r, y;
 
 	row = pane_row_get(p, pos);
 
-	y = p->y + (pos % p->height); /* relative position on screen */
 	cursorsave();
-	cursormove(p->x, y);
+	cursormove(p->x, p->y + (pos % p->height));
 
-	r = 0;
-	if (pos == p->pos) {
-		attrmode(ATTR_REVERSE_ON);
-		if (!p->focused)
-			attrmode(ATTR_FAINT_ON);
-		r = 1;
-	} else if (p->nrows && pos < p->nrows && row && row->bold) {
-		attrmode(ATTR_BOLD_ON);
-		r = 1;
-	}
+	if (p->focused)
+		THEME_ITEM_FOCUS();
+	else
+		THEME_ITEM_NORMAL();
+	if (row && row->bold)
+		THEME_ITEM_BOLD();
+	if (selected)
+		THEME_ITEM_SELECTED();
 	if (row)
 		printpad(pane_row_text(p, row), p->width);
 	else
-		printf("%-*.*s", p->width, p->width, "");
-	if (r)
-		attrmode(ATTR_RESET);
+		ttywritef("%-*.*s", p->width, p->width, "");
+
+	attrmode(ATTR_RESET);
 	cursorrestore();
 }
 
-void
-pane_row_redraw(struct pane *p, off_t pos)
-{
-	off_t prev;
-
-	prev = p->pos;
-	p->pos = pos;
-	pane_row_draw(p, prev); /* draw previous row again */
-	pane_row_draw(p, p->pos); /* draw new highlighted row */
-	fflush(stdout); /* flush and update directly */
-}
-
 void
 pane_setpos(struct pane *p, off_t pos)
 {
@@ -658,28 +676,16 @@ pane_setpos(struct pane *p, off_t pos)
 	if (pos == p->pos)
 		return; /* no change */
 
-	/* is on different scroll region? mark dirty */
+	/* is on different scroll region? mark whole pane dirty */
 	if (((p->pos - (p->pos % p->height)) / p->height) !=
 	    ((pos - (pos % p->height)) / p->height)) {
-		p->pos = pos;
 		p->dirty = 1;
-		pane_draw(p);
 	} else {
-		/* only redraw the 1 or 2 dirty rows */
-		pane_row_redraw(p, pos);
+		/* only redraw the 2 dirty rows */
+		pane_row_draw(p, p->pos, 0);
+		pane_row_draw(p, pos, 1);
 	}
-}
-
-void
-pane_setstartpos(struct pane *p)
-{
-	pane_setpos(p, 0);
-}
-
-void
-pane_setendpos(struct pane *p)
-{
-	pane_setpos(p, p->nrows - 1);
+	p->pos = pos;
 }
 
 void
@@ -723,14 +729,15 @@ pane_draw(struct pane *p)
 	if (p->hidden || !p->dirty)
 		return;
 
+	/* draw visible rows */
 	pos = p->pos - (p->pos % p->height);
 	for (y = 0; y < p->height; y++)
-		pane_row_draw(p, y + pos);
+		pane_row_draw(p, y + pos, (y + pos) == p->pos);
 
 	p->dirty = 0;
 }
 
-/* cycle visible pane in a direction, but don't cycle back */
+/* Cycle visible pane in a direction, but don't cycle back. */
 void
 cyclepanen(int n)
 {
@@ -754,7 +761,7 @@ cyclepanen(int n)
 	}
 }
 
-/* cycle visible panes */
+/* Cycle visible panes. */
 void
 cyclepane(void)
 {
@@ -774,7 +781,8 @@ updategeom(void)
 
 	panes[PaneFeeds].x = 0;
 	panes[PaneFeeds].y = 0;
-	panes[PaneFeeds].height = win.height - 1; /* space for statusbar */
+	/* reserve space for statusbar */
+	panes[PaneFeeds].height = win.height > 1 ? win.height - 1 : 1;
 
 	/* NOTE: updatesidebar() must happen before this function for the
 	   remaining width */
@@ -789,10 +797,10 @@ updategeom(void)
 		x = 0;
 	}
 
-	panes[PaneItems].width = w - 1; /* rest and space for scrollbar */
-	panes[PaneItems].height = win.height - 1; /* space for statusbar */
 	panes[PaneItems].x = x;
-	panes[PaneItems].y = 0;
+	panes[PaneItems].width = w > 0 ? w - 1 : 0; /* rest and space for scrollbar */
+	panes[PaneItems].height = panes[PaneFeeds].height;
+	panes[PaneItems].y = panes[PaneFeeds].y;
 
 	scrollbars[PaneFeeds].x = panes[PaneFeeds].x + panes[PaneFeeds].width;
 	scrollbars[PaneFeeds].y = panes[PaneFeeds].y;
@@ -806,7 +814,7 @@ updategeom(void)
 	/* statusbar below */
 	statusbar.width = win.width;
 	statusbar.x = 0;
-	statusbar.y = win.height - 1;
+	statusbar.y = panes[PaneFeeds].height;
 
 	alldirty();
 }
@@ -854,20 +862,30 @@ scrollbar_draw(struct scrollbar *s)
 		return;
 
 	cursorsave();
-	if (!s->focused)
-		attrmode(ATTR_FAINT_ON);
+
+	/* draw bar (not tick) */
+	if (s->focused)
+		THEME_SCROLLBAR_FOCUS();
+	else
+		THEME_SCROLLBAR_NORMAL();
 	for (y = 0; y < s->size; y++) {
+		if (y >= s->tickpos && y < s->tickpos + s->ticksize)
+			continue; /* skip tick */
 		cursormove(s->x, s->y + y);
-		if (y >= s->tickpos && y < s->tickpos + s->ticksize) {
-			attrmode(ATTR_REVERSE_ON);
-			fputs(" ", stdout);
-			attrmode(ATTR_REVERSE_OFF);
-		} else {
-			fputs(SCROLLBAR_SYMBOL, stdout);
-		}
+		ttywrite(SCROLLBAR_SYMBOL_BAR);
 	}
-	if (!s->focused)
-		attrmode(ATTR_RESET);
+
+	/* draw tick */
+	if (s->focused)
+		THEME_SCROLLBAR_TICK_FOCUS();
+	else
+		THEME_SCROLLBAR_TICK_NORMAL();
+	for (y = s->tickpos; y < s->size && y < s->tickpos + s->ticksize; y++) {
+		cursormove(s->x, s->y + y);
+		ttywrite(SCROLLBAR_SYMBOL_TICK);
+	}
+
+	attrmode(ATTR_RESET);
 	cursorrestore();
 	s->dirty = 0;
 }
@@ -953,15 +971,17 @@ uiprompt(int x, int y, char *fmt, ...)
 
 	cursorsave();
 	cursormove(x, y);
-	attrmode(ATTR_REVERSE_ON);
-	fputs(buf, stdout);
-	attrmode(ATTR_REVERSE_OFF);
+	THEME_INPUT_LABEL();
+	ttywrite(buf);
+	attrmode(ATTR_RESET);
+
+	THEME_INPUT_NORMAL();
 	cleareol();
 	cursormode(1);
 	cursormove(x + colw(buf) + 1, y);
-	fflush(stdout);
 
 	input = lineeditor();
+	attrmode(ATTR_RESET);
 
 	cursormode(0);
 	cursorrestore();
@@ -977,9 +997,9 @@ statusbar_draw(struct statusbar *s)
 
 	cursorsave();
 	cursormove(s->x, s->y);
-	attrmode(ATTR_REVERSE_ON);
+	THEME_STATUSBAR();
 	printpad(s->text, s->width);
-	attrmode(ATTR_REVERSE_OFF);
+	attrmode(ATTR_RESET);
 	cursorrestore();
 	s->dirty = 0;
 }
@@ -995,7 +1015,7 @@ statusbar_update(struct statusbar *s, const char *text)
 	s->dirty = 1;
 }
 
-/* line to item, modifies and splits line in-place */
+/* Line to item, modifies and splits line in-place. */
 int
 linetoitem(char *line, struct item *item)
 {
@@ -1108,7 +1128,6 @@ feed_load(struct feed *f, FILE *fp)
 	static struct item *items = NULL;
 	static size_t nitems = 0;
 	struct pane *p;
-	struct row *row;
 	size_t i;
 
 	for (i = 0; i < nitems; i++) {
@@ -1116,25 +1135,17 @@ feed_load(struct feed *f, FILE *fp)
 		free(items[i].link);
 	}
 	free(items);
-	items = NULL;
-	nitems = 0;
-
-	p = &panes[PaneItems];
-	free(p->rows);
-	p->rows = NULL;
-	p->nrows = 0;
 
 	if (feed_getitems(f, fp, &items, &nitems) == -1)
 		err(1, "%s: %s", __func__, f->name);
 
+	p = &panes[PaneItems];
 	p->pos = 0;
 	p->nrows = nitems;
+	free(p->rows);
 	p->rows = ecalloc(sizeof(p->rows[0]), nitems + 1);
-	for (i = 0; i < nitems; i++) {
-		row = &(p->rows[i]); /* do not use pane_row_get */
-		row->text = ""; /* custom formatter */
-		row->data = &items[i];
-	}
+	for (i = 0; i < nitems; i++)
+		p->rows[i].data = &items[i]; /* do not use pane_row_get */
 
 	updatenewitems(f);
 
@@ -1177,7 +1188,7 @@ feed_setenv(struct feed *f)
 		unsetenv("SFEED_FEED_PATH");
 }
 
-/* change feed, have one file open, reopen file if needed */
+/* Change feed, have one file open, reopen file if needed. */
 void
 feeds_set(struct feed *f)
 {
@@ -1301,7 +1312,6 @@ updatesidebar(int onlynew)
 		feed = &feeds[i];
 
 		row = &(p->rows[nrows]);
-		row->text = ""; /* custom formatter is used */
 		row->bold = (feed->totalnew > 0);
 		row->data = feed;
 
@@ -1354,25 +1364,12 @@ draw(void)
 	struct item *item;
 	size_t i;
 
-	if (win.width <= 1 || win.height <= 3 ||
-	    (!panes[PaneFeeds].hidden && win.width <= panes[PaneFeeds].width + 2))
-		return;
-
 	if (win.dirty) {
 		clearscreen();
 		win.dirty = 0;
 	}
 
-	/* if item selection text changed, update the status text */
-	if (panes[PaneItems].nrows &&
-	    (row = pane_row_get(&panes[PaneItems], panes[PaneItems].pos))) {
-		item = (struct item *)row->data;
-		statusbar_update(&statusbar, item->fields[FieldLink]);
-	} else {
-		statusbar_update(&statusbar, "");
-	}
-
-	/* NOTE: theres the same amount and indices of panes and scrollbars. */
+	/* There is the same amount and indices of panes and scrollbars. */
 	for (i = 0; i < LEN(panes); i++) {
 		pane_setfocus(&panes[i], i == selpane);
 		pane_draw(&panes[i]);
@@ -1383,8 +1380,15 @@ draw(void)
 		                 panes[i].nrows, panes[i].height);
 		scrollbar_draw(&scrollbars[i]);
 	}
+
+	/* If item selection text changed then update the status text. */
+	if ((row = pane_row_get(&panes[PaneItems], panes[PaneItems].pos))) {
+		item = (struct item *)row->data;
+		statusbar_update(&statusbar, item->fields[FieldLink]);
+	} else {
+		statusbar_update(&statusbar, "");
+	}
 	statusbar_draw(&statusbar);
-	fflush(stdout);
 }
 
 void
@@ -1420,9 +1424,9 @@ mousereport(int button, int release, int x, int y)
 			if (!p->nrows || pos >= p->nrows)
 				break;
 			pane_setpos(p, pos);
-			if (i == PaneFeeds) {
+			if (selpane == PaneFeeds) {
 				readurls();
-				row = pane_row_get(p, pos);
+				row = pane_row_get(p, p->pos);
 				f = (struct feed *)row->data;
 				feeds_set(f);
 				if (f->fp)
@@ -1430,12 +1434,12 @@ mousereport(int button, int release, int x, int y)
 				/* redraw row: counts could be changed */
 				updatesidebar(onlynew);
 				updatetitle();
-			} else if (i == PaneItems) {
+			} else if (selpane == PaneItems) {
 				if (dblclick && !changedpane) {
-					row = pane_row_get(&panes[PaneItems], pos);
+					row = pane_row_get(p, p->pos);
 					item = (struct item *)row->data;
 					markread(p, p->pos, p->pos, 1);
-					plumb(plumber, item->fields[FieldLink]);
+					forkexec((char *[]) { plumber, item->fields[FieldLink], NULL });
 				}
 			}
 			break;
@@ -1443,12 +1447,11 @@ mousereport(int button, int release, int x, int y)
 			if (!p->nrows || pos >= p->nrows)
 				break;
 			pane_setpos(p, pos);
-			if (i == PaneItems) {
-				p = &panes[PaneItems];
+			if (selpane == PaneItems) {
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
 				markread(p, p->pos, p->pos, 1);
-				pipeitem(piper, item, 1);
+				pipeitem(piper, item, -1, 1);
 			}
 			break;
 		case 3: /* scroll up */
@@ -1459,7 +1462,7 @@ mousereport(int button, int release, int x, int y)
 	}
 }
 
-/* custom formatter for feed row */
+/* Custom formatter for feed row. */
 char *
 feed_row_format(struct pane *p, struct row *row)
 {
@@ -1472,8 +1475,10 @@ feed_row_format(struct pane *p, struct row *row)
 
 	len = snprintf(counts, sizeof(counts), "(%lu/%lu)",
 	               feed->totalnew, feed->total);
-	utf8pad(bufw, sizeof(bufw), feed->name, p->width - len, ' ');
-	snprintf(text, sizeof(text), "%s%s", bufw, counts);
+	if (utf8pad(bufw, sizeof(bufw), feed->name, p->width - len, ' ') != -1)
+		snprintf(text, sizeof(text), "%s%s", bufw, counts);
+	else
+		text[0] = '\0';
 
 	return text;
 }
@@ -1488,6 +1493,7 @@ feed_row_match(struct pane *p, struct row *row, const char *s)
 	return (strcasestr(feed->name, s) != NULL);
 }
 
+#ifdef LAZYLOAD
 struct row *
 item_row_get(struct pane *p, off_t pos)
 {
@@ -1520,8 +1526,9 @@ item_row_get(struct pane *p, off_t pos)
 	}
 	return itemrow;
 }
+#endif
 
-/* custom formatter for item row */
+/* Custom formatter for item row. */
 char *
 item_row_format(struct pane *p, struct row *row)
 {
@@ -1537,7 +1544,6 @@ item_row_format(struct pane *p, struct row *row)
 		         tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
 		         tm.tm_hour, tm.tm_min, item->fields[FieldTitle]);
 	} else {
-		localtime_r(&(item->timestamp), &tm);
 		snprintf(text, sizeof(text), "%c                  %s",
 		         item->fields[FieldEnclosure][0] ? '@' : ' ',
 		         item->fields[FieldTitle]);
@@ -1554,7 +1560,7 @@ markread(struct pane *p, off_t from, off_t to, int isread)
 	FILE *fp;
 	off_t i;
 	const char *cmd;
-	int changed, isnew = !isread, pid, wpid, status;
+	int isnew = !isread, pid, wpid, status, visstart;
 
 	if (!urlfile || !p->nrows)
 		return;
@@ -1579,7 +1585,7 @@ markread(struct pane *p, off_t from, off_t to, int isread)
 			err(1, "popen");
 
 		for (i = from; i <= to; i++) {
-			row = &(p->rows[i]);
+			row = &(p->rows[i]); /* use pane_row_get: no need for lazyload */
 			item = (struct item *)row->data;
 			if (item->isnew != isnew) {
 				fputs(item->link, fp);
@@ -1596,20 +1602,20 @@ markread(struct pane *p, off_t from, off_t to, int isread)
 		/* fail: exit statuscode was non-zero */
 		if (status)
 			break;
-		for (i = from, changed = 0; i <= to && i < p->nrows; i++) {
+
+		visstart = p->pos - (p->pos % p->height); /* visible start */
+		for (i = from; i <= to && i < p->nrows; i++) {
 			row = &(p->rows[i]);
 			item = (struct item *)row->data;
-			if (item->isnew != isnew) {
-				row->bold = item->isnew = isnew;
-				curfeed->totalnew += isnew ? 1 : -1;
-				changed++;
-			}
-		}
-		if (changed) {
-			if (from != to || from != p->pos)
-				p->dirty = 1;
-			else
-				pane_row_redraw(p, from);
+			if (item->isnew == isnew)
+				continue;
+
+			row->bold = item->isnew = isnew;
+			curfeed->totalnew += isnew ? 1 : -1;
+
+			/* draw if visible on screen */
+			if (i >= visstart && i < visstart + p->height)
+				pane_row_draw(p, i, i == p->pos);
 		}
 		updatesidebar(onlynew);
 		updatetitle();
@@ -1673,18 +1679,27 @@ main(int argc, char *argv[])
 	int ch, button, fd, x, y, release;
 	off_t off;
 
+#ifdef __OpenBSD__
+	if (pledge("stdio rpath tty proc exec", NULL) == -1)
+		err(1, "pledge");
+#endif
+
 	setlocale(LC_CTYPE, "");
 
 	if ((tmp = getenv("SFEED_PLUMBER")))
 		plumber = tmp;
 	if ((tmp = getenv("SFEED_PIPER")))
 		piper = tmp;
+	if ((tmp = getenv("SFEED_YANKER")))
+		yanker = tmp;
 	urlfile = getenv("SFEED_URL_FILE");
 
 	panes[PaneFeeds].row_format = feed_row_format;
 	panes[PaneFeeds].row_match = feed_row_match;
 	panes[PaneItems].row_format = item_row_format;
+#ifdef LAZYLOAD
 	panes[PaneItems].row_get = item_row_get;
+#endif
 
 	feeds = ecalloc(argc, sizeof(struct feed));
 	if (argc == 1) {
@@ -1723,7 +1738,7 @@ main(int argc, char *argv[])
 		selpane = PaneItems;
 	}
 
-	if ((devnullfd = open("/dev/null", O_WRONLY)) < 0)
+	if ((devnullfd = open("/dev/null", O_WRONLY)) == -1)
 		err(1, "open: /dev/null");
 
 	updatesidebar(onlynew);
@@ -1816,11 +1831,13 @@ keyright:
 			break;
 startpos:
 		case 'g':
-			pane_setstartpos(&panes[selpane]);
+			pane_setpos(&panes[selpane], 0);
 			break;
 endpos:
 		case 'G':
-			pane_setendpos(&panes[selpane]);
+			p = &panes[selpane];
+			if (p->nrows)
+				pane_setpos(p, p->nrows - 1);
 			break;
 prevpage:
 		case 2: /* ^B */
@@ -1877,11 +1894,11 @@ nextpage:
 		case 'a': /* attachment */
 		case 'e': /* enclosure */
 		case '@':
-			if (selpane == PaneItems && panes[PaneItems].nrows) {
-				p = &panes[PaneItems];
+			if (selpane == PaneItems && panes[selpane].nrows) {
+				p = &panes[selpane];
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
-				plumb(plumber, item->fields[FieldEnclosure]);
+				forkexec((char *[]) { plumber, item->fields[FieldEnclosure], NULL });
 			}
 			break;
 		case 'm': /* toggle mouse mode */
@@ -1890,7 +1907,7 @@ nextpage:
 			break;
 		case 's': /* toggle sidebar */
 			panes[PaneFeeds].hidden = !panes[PaneFeeds].hidden;
-			if (selpane == PaneFeeds && panes[PaneFeeds].hidden)
+			if (selpane == PaneFeeds && panes[selpane].hidden)
 				selpane = PaneItems;
 			updategeom();
 			break;
@@ -1902,7 +1919,7 @@ nextpage:
 		case 'o': /* feeds: load, items: plumb url */
 		case '\n':
 			p = &panes[selpane];
-			if (selpane == PaneFeeds && panes[PaneFeeds].nrows) {
+			if (selpane == PaneFeeds && panes[selpane].nrows) {
 				readurls();
 				row = pane_row_get(p, p->pos);
 				f = (struct feed *)row->data;
@@ -1912,28 +1929,28 @@ nextpage:
 				/* redraw row: counts could be changed */
 				updatesidebar(onlynew);
 				updatetitle();
-			} else if (selpane == PaneItems && panes[PaneItems].nrows) {
+			} else if (selpane == PaneItems && panes[selpane].nrows) {
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
 				markread(p, p->pos, p->pos, 1);
-				plumb(plumber, item->fields[FieldLink]);
+				forkexec((char *[]) { plumber, item->fields[FieldLink], NULL });
 			}
 			break;
 		case 'c': /* items: pipe TSV line to program */
 		case 'p':
 		case '|':
-		case 'y': /* yank: pipe TSV line to yank url to clipboard */
-		case 'E': /* yank: pipe TSV line to yank enclosure to clipboard */
-			if (selpane == PaneItems && panes[PaneItems].nrows) {
+		case 'y': /* yank: pipe TSV field to yank url to clipboard */
+		case 'E': /* yank: pipe TSV field to yank enclosure to clipboard */
+			if (selpane == PaneItems && panes[selpane].nrows) {
 				p = &panes[selpane];
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
 				switch (ch) {
-				case 'y': pipeitem("cut -f 3 | xclip -r", item, 0); break;
-				case 'E': pipeitem("cut -f 8 | xclip -r", item, 0); break;
+				case 'y': pipeitem(yanker, item, FieldLink, 0); break;
+				case 'E': pipeitem(yanker, item, FieldEnclosure, 0); break;
 				default:
 					markread(p, p->pos, p->pos, 1);
-					pipeitem(piper, item, 1);
+					pipeitem(piper, item, -1, 1);
 					break;
 				}
 			}
@@ -1947,7 +1964,7 @@ nextpage:
 			break;
 		case 'r': /* mark item as read */
 		case 'u': /* mark item as unread */
-			if (selpane == PaneItems && panes[PaneItems].nrows) {
+			if (selpane == PaneItems && panes[selpane].nrows) {
 				p = &panes[selpane];
 				markread(p, p->pos, p->pos, ch == 'r');
 			}
